package dareka.processor;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.Socket;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import dareka.common.CloseUtil;
import dareka.common.Config;
import dareka.common.ConfigObserver;
import dareka.common.FileUtil;
import dareka.common.Logger;

/**
 *  [nl] [Jt@C\[X
 * @since NicoCache_nl+101219mod
 */
public class LocalFileResource extends Resource implements ConfigObserver {
	private File file;
	private long start, end, decodedLength;
	private String statusline;
	private boolean clientCanKeepAlive;
	private boolean needsSendingBody, needsDecodingBody;
	private boolean setLastModified = true;
	private HttpResponseHeader responseHeader;
	
	private static final Map<String, String> mimeTypes =
		 Collections.synchronizedMap(new LinkedHashMap<String, String>());
	private static File mimeTypesFile;
	private static volatile long mimeTypesLastModified;
	
	static {
		Config.addObserver(new LocalFileResource());
	}
	private LocalFileResource() {}
	
	/**
	 * [Jt@C\[X̃CX^X𐶐B
	 * 
	 * @param file [Jt@C
	 * @throws IOException o̓G[
	 */
	public LocalFileResource(File file) throws IOException {
		if (file == null) {
			throw new IllegalArgumentException("file must not be null");
		}
		if (Boolean.getBoolean("localPathCheck")) {
			// JgfBNg܂܂ȂpX͋֎~
			String cwd = new File(".").getCanonicalPath() + File.separator;
			String path = file.getCanonicalPath();
			if (!path.startsWith(cwd)) {
				throw new SecurityException("Invalid local path: " + path);
			}
		}
		if (!file.exists()) {
			File gzFile = new File(file.getPath() + ".gz");
			if (gzFile.exists()) {
				file = gzFile;
			}
		}
		this.file = file;
		this.end = file.length() - 1L;
	}
	
	/**
	 * file = new File(path)
	 * @param path [Jt@C̃pX
	 * @see #LocalFileResource(File)
	 */
	public LocalFileResource(String path) throws IOException {
		this(new File(path));
	}
	
	/**
	 * @param setLastModified Last-Modifiedwb_t邩H
	 * @see #LocalFileResource(File)
	 */
	public LocalFileResource(File file, boolean setLastModified)
			throws IOException {
		this(file);
		this.setLastModified = setLastModified;
	}
	
	public File getFile() {
		return file;
	}
	
	public long getLength() {
		if (needsDecodingBody) {
			return getDecodedLength();
		}
		return end - start + 1L;
	}
	
	/**
	 * t@C]\H(t@C݂āAǂݍ݉\)
	 * 
	 * @return t@C]\Ȃtrue
	 */
	public boolean isValid() {
		return file.isFile() && file.canRead();
	}
	
	/**
	 * X|Xwb_擾B
	 * 
	 * @param requestHeader NGXgwb_(null ȂftHgwb_p)
	 * @param setMandatoryHeader trueȂK{wb_(Content-Length, Keep-Alive)t
	 * @return X|Xwb_
	 * @throws IOException X|Xwb_̐Ɏs
	 */
	public HttpResponseHeader getResponseHeader(HttpRequestHeader requestHeader,
			boolean setMandatoryHeader) throws IOException {
		needsSendingBody = needsDecodingBody = false;
		
		String contentEncoding = null;
		String path = file.getPath();
		if (requestHeader == null) {
			requestHeader = new HttpRequestHeader(HttpHeader.GET + " " +
					"LOCAL/" + path + " HTTP/1.1\r\n\r\n");
		}
		if (path.endsWith(".gz")) {
			path = path.substring(0, path.length() - 3);
			contentEncoding = HttpHeader.CONTENT_ENCODING_GZIP;
			if (!requestHeader.isGzipAccepted()) {
				needsDecodingBody = true;
			}
		}
		
		if (isValid()) {
			statusline = "HTTP/1.1 200 OK";
			String method = requestHeader != null ?
					requestHeader.getMethod() : HttpHeader.GET;
			if (HttpHeader.GET.equalsIgnoreCase(method)) {
				if (checkIfModifiedSince(requestHeader)) {
					statusline = "HTTP/1.1 304 Not Modified";
				} else if (checkIfPartialContent(requestHeader)){
					needsSendingBody = true;
					statusline = "HTTP/1.1 206 Partial Content";
				} else {
					needsSendingBody = true;
				}
			} else if (HttpHeader.HEAD.equalsIgnoreCase(method)) {
				// do nothing.
			} else if (HttpHeader.POST.equalsIgnoreCase(method)) {
				statusline = "HTTP/1.1 405 Method Not Allowed";
			} else {
				statusline = "HTTP/1.1 501 Not Implemented";
			}
		} else {
			statusline = "HTTP/1.1 404 Not Found";
		}
		
		responseHeader = new HttpResponseHeader(statusline + "\r\n\r\n");
		if (contentEncoding != null) {
			responseHeader.setMessageHeader(HttpHeader.CONTENT_ENCODING,
					contentEncoding);
		}
		String contentType = getMimeType(path);
		if (contentType != null) {
			responseHeader.setMessageHeader(HttpHeader.CONTENT_TYPE,
					contentType);
		}
		clientCanKeepAlive = isClientCanKeepAlive(requestHeader);
		if (setMandatoryHeader) {
			doSetMandatoryResponseHeader(responseHeader);
		}
		return responseHeader;
	}
	
	/**
	 * X|X{fB擾BɃX|Xwb_擾KvB
	 * X|Xwb_̐ʂɂĂ̓X|X{fB̏ꍇB
	 * 
	 * @return X|X{fB(̏ꍇ̓TCY0̃oCgz)
	 */
	public byte[] getResponseBody() {
		InputStream in = null;
		try {
			if (needsSendingBody) {
				in = getDecodedInputStream();
				ByteArrayOutputStream bout = new ByteArrayOutputStream();
				if (FileUtil.copy(in, bout) > 0) {
					return bout.toByteArray();
				}
			}
		} catch (IOException e) {
			Logger.error(e);
		} finally {
			CloseUtil.close(in);
		}
		return new byte[0];
	}
	
	private InputStream getDecodedInputStream() throws IOException {
		InputStream in = new FileInputStream(file);
		String contentEncoding = null;
		if (responseHeader != null) {
			contentEncoding = responseHeader.getMessageHeader(
					HttpHeader.CONTENT_ENCODING);
		}
		return HttpUtil.getDecodedInputStream(in, contentEncoding);
	}
	
	private long getDecodedLength() {
		if (decodedLength == 0L) {
			InputStream in = null;
			try {
				in = getDecodedInputStream();
				while (in.read() != -1) {
					decodedLength++;
				}
			} catch (IOException e) {
				Logger.error(e);
			} finally {
				CloseUtil.close(in);
			}
		}
		return decodedLength;
	}
	
	private byte[] getRangeBody() {
		FileInputStream in = null;
		try {
			if (needsSendingBody && getLength() <= Integer.MAX_VALUE) {
				in = new FileInputStream(file);
				in.skip(start);
				byte[] buf = new byte[(int) getLength()];
				int readed = 0;
				while (readed < buf.length) {
					readed += in.read(buf, readed, buf.length - readed);
				}
				return buf;
			}
		} catch (IOException e) {
			Logger.error(e);
		} finally {
			CloseUtil.close(in);
		}
		return new byte[0];
	}
	
	private boolean checkIfModifiedSince(HttpRequestHeader requestHeader) {
		if (requestHeader != null) {
			long lm = file.lastModified();
			String ims = requestHeader.getMessageHeader(HttpHeader.IF_MODIFIED_SINCE);
			if (ims != null && (lm - HttpHeader.parseDateString(ims)) < 1000L) {
				return true;
			}
		}
		return false;
	}
	
	private static final Pattern RANGE_PATTERN = Pattern.compile(
			"(?i)bytes\\s*=\\s*(\\d+)\\s*-\\s*(\\d+)?");
	
	private boolean checkIfPartialContent(HttpRequestHeader requestHeader) {
		if (requestHeader != null && !needsDecodingBody) {
			String range = requestHeader.getMessageHeader("Range");
			if (range != null) {
				Matcher m = RANGE_PATTERN.matcher(range);
				if (m.find()) {
					start = Integer.parseInt(m.group(1));
					if (m.group(2) != null) {
						int n = Integer.parseInt(m.group(2));
						if (start < n && n < end) {
							end = n;
						}
					}
					return true;
				}
			}
		}
		return false;
	}
	
	@Override
	protected boolean endEnsuredTransferTo(Socket receiver,
			HttpRequestHeader requestHeader, Config config) throws IOException {
		OutputStream out = receiver.getOutputStream();
		HttpResponseHeader rh = getResponseHeader(requestHeader, false);
		
		execSendingHeaderSequence(out, rh);
		
		if (needsSendingBody) {
			String contentType = rh.getMessageHeader(HttpHeader.CONTENT_TYPE);
			boolean useRange = getLength() < file.length();
			InputStream in = null;
			try {
				StringBuilder sb = new StringBuilder(file.getPath());
				if (useRange) {
					in = new ByteArrayInputStream(getRangeBody());
					sb.append(String.format("[range=%d-%d]", start, end));
				} else if (needsDecodingBody) {
					in = getDecodedInputStream();
				} else {
					in = new FileInputStream(file);
				}
				sb.append(" ").append(contentType);
				Logger.debugWithThread(sb.toString());
				execSendingBodySequence(out, in, rh.getContentLength());
			} finally {
				CloseUtil.close(in);
			}
		}
		return clientCanKeepAlive;
	}
	
	@Override
	protected void doSetMandatoryResponseHeader(HttpResponseHeader responseHeader) {
		long contentLength = 0L;
		if (isValid()) {
			if (needsSendingBody) {
				contentLength = getLength();
			}
			if (setLastModified) {
				responseHeader.setMessageHeader(HttpHeader.LAST_MODIFIED,
						HttpHeader.getDateString(file.lastModified()));
			}
		}
		responseHeader.setContentLength(contentLength);
		responseHeader.setMessageHeader(HttpHeader.DATE,
				HttpHeader.getDateString(System.currentTimeMillis()));
		
		if (responseHeader.getStatusCode() != 200) {
			responseHeader.removeMessageHeader("Cache-Control");
			responseHeader.removeMessageHeader("Expires");
			responseHeader.removeMessageHeader("Pragma");
		}
		if (responseHeader.getStatusCode() == 206) {
			responseHeader.setMessageHeader("Accept-Ranges", "bytes");
			responseHeader.setMessageHeader("Content-Range",
					String.format("bytes %d-%d/%d", start, end, file.length()));
		}
		
		if (clientCanKeepAlive) {
			responseHeader.setMessageHeader(HttpHeader.CONNECTION,
					HttpHeader.CONNECTION_KEEP_ALIVE);
		} else {
			responseHeader.setMessageHeader(HttpHeader.CONNECTION,
					HttpHeader.CONNECTION_CLOSE);
		}
	}
	
	/**
	 * t@C̊gqɑΉ MimeType 擾B
	 * 啶Ə͋ʂ͂ȂB
	 * 
	 * @param path 擾Ώۂ̃t@C̃pX
	 * @return ΉMimeTypeA݂Ȃꍇnull
	 * @since NicoCache_nl+110219mod
	 */
	public static String getMimeType(String path) {
		if (path == null || path.length() == 0) {
			return null;
		}
		loadMimeTypes();
		int pos = path.lastIndexOf('.');
		if (pos != -1) {
			String ext = path.substring(pos + 1).toLowerCase();
			return mimeTypes.get(ext);
		}
		return null;
	}
	
	/**
	 * gq(key)MimeType(value)̓ǂݎp}bv擾B
	 * 
	 * @return gqMimeTypȇΉ}bvA݂Ȃꍇ͋}bv
	 * @since NicoCache_nl+110219mod
	 */
	public static Map<String, String> getMimeTypes() {
		loadMimeTypes();
		return Collections.unmodifiableMap(mimeTypes);
	}
	
	public void update(Config config) {
		String mimeTypesPath = System.getProperty("mimeTypes", "");
		if (mimeTypesPath.length() > 0) {
			mimeTypesFile = new File(mimeTypesPath);
			if (!mimeTypesFile.exists()) {
				File defaultFile = new File(mimeTypesPath + ".default");
				if (defaultFile.exists()) {
					mimeTypesFile = defaultFile;
				}
			}
		} else {
			mimeTypesFile = null;
			clearMimeTypes();
		}
	}
	
	private static void loadMimeTypes() {
		if (mimeTypesFile == null) {
			return;
		}
		long lm = mimeTypesFile.lastModified();
		if (mimeTypesLastModified < lm) {
			synchronized (mimeTypes) {
				if (mimeTypesLastModified == lm) {
					return; // ̃XbhŊɃ[hς
				}
				Logger.debugWithThread("loading " + mimeTypesFile.getPath());
				BufferedReader br = null;
				try {
					//br = new BufferedReader(new FileReader(mimeTypesFile));
					br = new BufferedReader(new InputStreamReader(
							new FileInputStream(mimeTypesFile), "UTF-8"));
					parseMimeTypes(br);
				} catch (IOException e) {
					Logger.error(e);
				} finally {
					CloseUtil.close(br);
				}
				mimeTypesLastModified = lm;
			}
		} else if (lm == 0L) {
			clearMimeTypes();
		}
	}
	
	private static void clearMimeTypes() {
		mimeTypes.clear();
		mimeTypesLastModified = 0L;
	}
	
	private static final Pattern MIME_LINE_PATTERN = Pattern.compile(
			"^([^#\\r\\n\\t]+)\\t++(.+)$");
	
	private static void parseMimeTypes(BufferedReader br) throws IOException {
		Map<String, String> newMimeTypes = new LinkedHashMap<String, String>();
		String line;
		while ((line = br.readLine()) != null) {
			Matcher m = MIME_LINE_PATTERN.matcher(line);
			if (!m.matches()) continue;
			String mimeType = m.group(1);
			for (String ext : m.group(2).split("\\s+")) {
				if (ext.startsWith(".")) ext = ext.substring(1);
				newMimeTypes.put(ext, mimeType);
			}
		}
		mimeTypes.clear();
		mimeTypes.putAll(newMimeTypes);
	}
	
}
