package dareka.processor.impl;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.concurrent.FutureTask;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import dareka.common.CloseUtil;
import dareka.common.DiskFreeSpace;
import dareka.common.Logger;
import dareka.extensions.CompleteCache;
import dareka.extensions.SystemEventListener;
import dareka.processor.HttpResponseHeader;
import dareka.processor.TransferListener;

public class NicoCachingListener implements TransferListener {
    private static final int BUF_SIZE = 32 * 1024;
    private static final Pattern CONTENT_RANGE_VALUE_PATTERN =
            Pattern.compile("bytes (\\d+)-(\\d+)/(\\d+)");

    private Cache cache;
    private FutureTask<String> retrieveTitleTask;
    private FileChannel cacheChannel = null;
    private boolean errorOccured;
    private boolean keepCacheOnError = false;

    // [nl]
    InputStream cacheInput = null;
    int tmpSize = 0;
    boolean partial = false;
    boolean dupconnect = false;
    NLEventSource eventSource;

    public NicoCachingListener(Cache cache,
            FutureTask<String> retrieveTitleTask, InputStream cacheInput,
            boolean dupconnect, NLEventSource eventSource) {
        this.cache = cache;
        this.retrieveTitleTask = retrieveTitleTask;
        this.cacheInput = cacheInput;
        this.dupconnect = dupconnect;
        this.eventSource = eventSource;
        errorOccured = false;
    }

    // W[̓wb_200ɂ
    public void onResponseHeader(HttpResponseHeader responseHeader) {
        long contentLength = responseHeader.getContentLength();

        if (contentLength <= 0) {
            errorOccured = true;
        }

        int statusCode = responseHeader.getStatusCode();

        if (statusCode == 206) {
            String rangeValue =
                    responseHeader.getMessageHeader("Content-Range");
            Matcher m = CONTENT_RANGE_VALUE_PATTERN.matcher(rangeValue);
            if (m.find()) {
                if (Integer.parseInt(m.group(2)) + 1 == Integer.parseInt(m.group(3))) {
                    Logger.info("Partial download from " + m.group(1) + " byte");
                } else {
                    Logger.info("Maybe Bad Responce: " + m.group(0));
                }
                contentLength = Long.parseLong(m.group(3));

                // [nl] DL镔܂ł̃TCYݒ肵ALbVMtOON
                tmpSize = Integer.parseInt(m.group(1));
                Cache.setDLFlag(cache.getId(), tmpSize);
                partial = true;
            }
            // [nl] wb_C
            responseHeader.setStatusCode(200, "OK");
            responseHeader.removeMessageHeader("Content-Range");
            responseHeader.removeMessageHeader("Accept-Range");
            responseHeader.setContentLength(contentLength);
        } else if (statusCode == 416) {
        	Logger.warning("̃W[Ɏs܂BuEU[hĂB");
        	try {
        		CloseUtil.close(cacheInput);
        		cache.deleteTmp();
        	} catch (IOException e) {
        		Logger.error(e);
        	}
            errorOccured = true;
        } else if (statusCode != 200) {
        	// [nl] 403ƂLbVƍ̂
        	responseHeader.addCacheControlHeaders(0);
            CloseUtil.close(cacheInput);
            errorOccured = true;
        }

        //[nl] 󂫗eʂ̃`FbN(rc2)
        //	   	wlȉTCY+100kbȉŒ~
        long freeSize = DiskFreeSpace.get(Cache.getCacheDir());
        long needSize = Long.getLong("needFreeSpace") * 1024 * 1024;
        if (needSize != 0) {
        	if (freeSize < needSize
        				|| freeSize < contentLength - tmpSize + 100 * 1024) {
        		Logger.warning(String.format(
        				"ceʂ %,d oCgȂ̂ŃLbV܂B", freeSize));
        		errorOccured = true;
        	}
        }

        try {
            if (!errorOccured && !dupconnect) {
                FileOutputStream fOut = cache.getTmpOutputStream(partial);
                cacheChannel = fOut.getChannel();
                
                // ɓ̃TCYmۂ (sp1.06)
                if (Boolean.getBoolean("cacheAllocateFirst")) {
                    cacheChannel.write(ByteBuffer.wrap(new byte[1]), contentLength - 1);
                    cacheChannel.truncate(contentLength);
                    cacheChannel.position(tmpSize);
                }
            }
        } catch (IOException e) {
            Logger.warning(cache.getCacheFileName() + ": " + e.toString());
        }

        // [nl] DL̓̍ŏITCY
        if (!dupconnect) {
        	Cache.setDLFinalSize(cache.getId(), (int) contentLength);
        }
        if (eventSource != null) {
        	eventSource.setResponseHeader(responseHeader);
        }
    }

    // LbVɗL镪𑗐M
    public void onTransferBegin(OutputStream bout) {
        if (partial && cacheInput != null) {
            try {
                int len = 0;
                int rest = tmpSize;
                byte[] buf = new byte[BUF_SIZE];
                while ((len = cacheInput.read(buf)) != -1 && rest > 0) {
                    try {
                        bout.write(buf, 0, (len < rest) ? len : rest);
                    } catch (IOException e) {
                        // uEUւ݂̏ŃG[ɂȂꍇ
                        // LbV폜Ȃ
                        keepCacheOnError = true;
                        throw e;
                    }
                    rest -= len;
                }
            } catch (IOException e) {
                Logger.warning(cache.getCacheFileName() + ": " + e.toString());
                errorOccured = true;
            } finally {
                if (CloseUtil.close(cacheInput) == false) {
                    errorOccured = true;
                }
            }
        }
    }

    public void onTransferring(byte[] buf, int length) {
        if (errorOccured || cacheChannel == null) {
            return;
        }

        try {
            cacheChannel.write(ByteBuffer.wrap(buf, 0, length), tmpSize);
            tmpSize += length;
        } catch (IOException e) {
            Logger.warning(cache.getCacheFileName() + ": " + e.toString());
            errorOccured = true;
        }

        if (!dupconnect) {
        	Cache.setDLFlag(cache.getId(), tmpSize);
        }
    }

    public void onTransferEnd(boolean completed) {
        if (dupconnect) {
            return;
        }

        // [nl] ܂DLĂȂ؂ (sp1.06)
        if (Boolean.getBoolean("cacheAllocateFirst")) {
        	try {
        		if (cacheChannel != null) {
        			cacheChannel.truncate(tmpSize);
        		}
        	} catch (IOException e) { }
        }

        if (CloseUtil.close(cacheChannel) == false) {
            errorOccured = true;
        }

        try {
            Wrapupper w =
                    selectWrapupper(completed, errorOccured, keepCacheOnError,
                            cache, retrieveTitleTask, eventSource);
            w.wrapup();
        } catch (IOException e) {
            Logger.debugWithThread(e);
            Logger.warning(e.toString());
        }

        // [nl] DLtOāALbV
        // Cache#store()ザȂExtensionɕsôŒ
        Cache.setDLFlag(cache.getId(), -1);
    }

    /**
     * Select wrapping up way.
     *
     * This condition selecting is a little bit complex,
     * so make this part a method to be able to test alone.
     *
     * @param completed
     * @param aErrorOccured
     * @param aKeepCacheOnError
     * @param aCache
     * @param aRetrieveTitleTask
     * @return Wrapupper
     */
    static Wrapupper selectWrapupper(boolean completed, boolean aErrorOccured,
            boolean aKeepCacheOnError, Cache aCache,
            FutureTask<String> aRetrieveTitleTask, NLEventSource eventSource) {
        // [nl] LbV̓폜
        if (NLShared.INSTANCE.isInDeleteSet(aCache.getVideoId())) {
        	NLShared.INSTANCE.removeDeleteSet(aCache.getVideoId());
        	return new Deleter(aCache, aRetrieveTitleTask, eventSource);
        }

        if (aErrorOccured) {
            return new Cleanupper(aKeepCacheOnError, aCache, aRetrieveTitleTask);
        } else if (!completed) {
            if (Boolean.getBoolean("resumeDownload")) {
                // [nl] G[ȂAPɊĂȂȂ
                return new Suspender(aCache, aRetrieveTitleTask, eventSource);
            } else {
                return new Cleanupper(aKeepCacheOnError, aCache,
                        aRetrieveTitleTask);
            }
        } else { // completed
            return new Completer(aCache, aRetrieveTitleTask, eventSource);
        }
    }

    static interface Wrapupper {
        void wrapup() throws IOException;
    }

    static class Cleanupper implements Wrapupper {
        private boolean keepCacheOnError;
        private Cache cache;
        private FutureTask<String> retrieveTitleTask;

        Cleanupper(boolean keepCacheOnError, Cache cache,
                FutureTask<String> retrieveTitleTask) {
            this.keepCacheOnError = keepCacheOnError;
            this.cache = cache;
            this.retrieveTitleTask = retrieveTitleTask;
        }

        public void wrapup() throws IOException {
            if (!keepCacheOnError) {
                cache.deleteTmp();
                Logger.debugWithThread(cache.getCacheFileName() + " deleted");
            }

            if (retrieveTitleTask != null) {
                retrieveTitleTask.cancel(true);
            }
        }
    }

    static class Suspender implements Wrapupper {
        private Cache cache;
        private FutureTask<String> retrieveTitleTask;
        private NLEventSource eventSource;

        Suspender(Cache cache, FutureTask<String> retrieveTitleTask,
        		NLEventSource eventSource) {
            this.cache = cache;
            this.retrieveTitleTask = retrieveTitleTask;
            this.eventSource = eventSource;
        }

        public void wrapup() throws IOException {
            String title;
            try {
                if (retrieveTitleTask != null
                        && (title = retrieveTitleTask.get()) != null) {
                    cache.setDescribe(title);
                    cache.setTmpDescribe(title);
                }
            } catch (Exception e) {
                Logger.warning("title retrieving failed: " + e.toString());
            }

            // [nl] f̃Cxgʒm
            if (eventSource != null) {
            	NLShared.INSTANCE.notifySystemEvent(
            			SystemEventListener.CACHE_SUSPENDED, eventSource, false);
            	if (!cache.existsTmp()) {
            		Logger.debug("no temporary cache exists: " + cache.getId());
            		return;
            	}
            }

            Logger.info("suspended: " + cache.getCacheFileName());
        }
    }

    static class Completer implements Wrapupper {
        private Cache cache;
        private FutureTask<String> retrieveTitleTask;
        private NLEventSource eventSource;

        Completer(Cache cache, FutureTask<String> retrieveTitleTask,
        		NLEventSource eventSource) {
            this.cache = cache;
            this.retrieveTitleTask = retrieveTitleTask;
            this.eventSource = eventSource;
        }

        public void wrapup() throws IOException {
            String title;
            try {
                if (retrieveTitleTask != null
                        && (title = retrieveTitleTask.get()) != null) {
                    cache.setDescribe(title);
                }
            } catch (Exception e) {
                Logger.warning("title retrieving failed: " + e.toString());
            }

            cache.store();

            // [nl] ݃`FbNƊ̃Cxgʒm
            if (!cache.exists()) {
            	Logger.debug("no complete cache exists: " + cache.getId());
            	return;
            }
            for (CompleteCache entry : NLShared.INSTANCE.getCompleteEntries()) {
            	try {
            		if (entry.onComplete(cache)) {
            			break; // trueԂĂ_ŏI
            		}
            	} catch (Throwable t) {
            		Logger.error(t); // G[͖đs
            	}
            }
            if (eventSource != null) {
            	NLShared.INSTANCE.notifySystemEvent(
            			SystemEventListener.CACHE_COMPLETED, eventSource, false);
            }

            Logger.info("cache completed: " + cache.getCacheFileName());
        }
    }

    // [nl] LbV̓폜p
    static class Deleter implements Wrapupper {
        private Cache cache;
        private FutureTask<String> retrieveTitleTask;
        private NLEventSource eventSource;

        Deleter(Cache cache, FutureTask<String> retrieveTitleTask,
        		NLEventSource eventSource) {
            this.cache = cache;
            this.retrieveTitleTask = retrieveTitleTask;
            this.eventSource = eventSource;
        }

        public void wrapup() throws IOException {
        	// LbVȂ̂ňꎞt@Ĉ܂
        	if (cache.deleteTmp()) {
        		if (eventSource != null) {
        			NLShared.INSTANCE.notifySystemEvent(
        					SystemEventListener.CACHE_REMOVED, eventSource, false);
        		}
        		Cache.removeAll(cache.getId(), false);
        		Logger.info("cache removed: " + cache.getCacheFileName());
        	} else {
        		Logger.info("cache removing failed: " + cache.getCacheFileName());
        	}
            if (retrieveTitleTask != null) {
                retrieveTitleTask.cancel(true);
            }
        }
    }

}
